/*
	PROforma, PROGS Font & Raster Manager

	PFPageScroll
		routine to scroll OR pan the page (not both)
		The PageBbox will internally be adjusted so that re-rendering the
		elements which build up the page will result in efficient redrawing.
		The distance to scroll/pan is specified in original user space (72pt/in)
		This routine should only be used to scroll in one direction at a time

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 20 July 1993
	19.10.1995 complete rework for different drivers etc
			   it now works for all drivers
	20.03.1996 PFPPageScroll added
*/

#include "gstate_h"

Error PFPageScroll(Gstate gstate, pt xdist, pt ydist)
{
	Error err;
	short xd,yd;					/* number of pixels to scroll/pan */

	/* scrolling in both directions together is invalid */
	if (xdist && ydist) return ERR_IEXP;

	return PFPPageScroll(gstate, pt2rshort(fixmul(xdist,gstate->todev.x)),
								 pt2rshort(fixmul(ydist,gstate->todev.y)) );
}

Error PFPPageScroll(Gstate gstate, short xd, short yd)
{
	Error err;

	/* scrolling in both directions together is invalid */
	if (xd && yd) return ERR_IEXP;

	gstate->Origin.x-=short2pt(xd);
	gstate->Origin.y-=short2pt(yd);

	/* handle scrolling over screen size or more */
	if ( abs(xd)>=gstate->raster->clip.xsiz ||
		 abs(yd)>=gstate->raster->clip.ysiz )
		return ERR_OK;

	/* actually scroll the contents of the page */
	err=gstate->bitmap->CopyArea(gstate->raster,gstate->raster,
			gstate->raster->clip.xsiz-abs(xd),
			gstate->raster->clip.ysiz-abs(yd),
			gstate->raster->clip.xorg + ( xd>0 ? 0 : -xd),
			gstate->raster->clip.yorg + ( yd>0 ? 0 : -yd),
			gstate->raster->clip.xorg + ( xd>0 ? xd : 0 ),
			gstate->raster->clip.yorg + ( yd>0 ? yd : 0 ));
	if (err) return ERR_OK;

	/* now adjust the PageBbox */
	if (xd>0)
	{
		gstate->raster->clip.xsiz=xd;
	}
	else if (xd<0)
	{
		xd=-xd; 			/* make value absolute */
		gstate->raster->clip.xorg+=gstate->raster->clip.xsiz-xd;
		gstate->Origin.x+=short2pt(gstate->raster->clip.xsiz-xd);
		gstate->raster->clip.xsiz=xd;
	}
	else if (yd>0)
	{
		gstate->raster->clip.ysiz=yd;
	}
	else if (yd<0)
	{
		yd=-yd; 			/* make value absolute */
		gstate->raster->clip.yorg+=gstate->raster->clip.ysiz-yd;
		gstate->Origin.y+=short2pt(gstate->raster->clip.ysiz-yd);
		gstate->raster->clip.ysiz=yd;
	}

	return ERR_OK;
}
