/* PROforma, PROGS FOnt & Raster MAnager

	Show and Width operators for strings.
	The backslash ('\') is interpretted as escape character. Between to
	backslashes you can give the name or code (decimal) of the character to
	print. A backslash can be included with the '\\' sequence.

   (c) PROGS, Professional & Graphical Software
   written by Joachim Van der Auwera
   started 14 April 1995
   */

#include "gstate_h"
#include "system_h"

Error PFStringShow(Gstate gstate, char *string)
{
	coor *w=&gstate->CharWidth;

	if (!string) return ERR_OK;

	while (StringContinues(gstate,*string))
	{
		PFShowChar(gstate, PFChar2UniCode(gstate,string,&string));
		PFMoveR(gstate, w->x, w->y);
	}
	return PFPathClear(gstate);
}

Error PFStringShowKern(Gstate gstate, char *string)
{
	coor *w=&gstate->CharWidth;
	short code, next;

	if (!string) return ERR_OK;

	code=PFChar2UniCode(gstate,string,&string);
	while (StringContinues(gstate,code))
	{
		PFShowChar(gstate, code);
		next=PFChar2UniCode(gstate,string,&string);
		KernAdjust(gstate, code, next);
		PFMoveR(gstate, w->x, w->y);
		code=next;
	}
	return PFPathClear(gstate);
}

Error PFStringShowX(Gstate gstate, char *string, pt *Xlist)
{
	if (!string) return ERR_OK;

	while (StringContinues(gstate,*string))
	{
		PFShowChar(gstate, PFChar2UniCode(gstate,string,&string));
		PFMoveR(gstate, *Xlist++, 0);
	}
	return PFPathClear(gstate);
}

Error PFStringShowXY(Gstate gstate, char *string, pt *Xlist, pt *Ylist)
{
	if (!string) return ERR_OK;

	while (StringContinues(gstate,*string))
	{
		PFShowChar(gstate, PFChar2UniCode(gstate,string,&string));
		PFMoveR(gstate, *Xlist++, *Ylist++);
	}
	return PFPathClear(gstate);
}

Error PFStringShowTrack(Gstate gstate, char *string, pt width)
{
	coor *w=&gstate->CharWidth;
	char *s=string;
	pt add=0, wdt; short len=0;

	if (!string) return ERR_OK;

	/* determine length of string */
	while (*s++) len++;
	PFStringWidth(gstate,string,&wdt);
	if (len) add=(width-wdt)/(len-1);

	while (StringContinues(gstate,*string))
	{
		PFShowChar(gstate, PFChar2UniCode(gstate,string,&string));
		PFMoveR(gstate, w->x+add, w->y);
	}
	return PFPathClear(gstate);
}

Error PFStringShowJust(Gstate gstate, char *string, pt width)
{
	coor *w=&gstate->CharWidth;
	char *s=string;
	pt add=0,wdt; short len=0;

	if (!string) return ERR_OK;

	/* determine length of string */
	while (*s) if (*s++==' ') len++;
	PFStringWidth(gstate,string,&wdt);
	if (len) add=(width-wdt)/len;

	while (StringContinues(gstate,*string))
	{
		PFShowChar(gstate, PFChar2UniCode(gstate,string,&string));
		if (*string++==' ')
			PFMoveR(gstate, w->x+add, w->y);
		else
			PFMoveR(gstate, w->x, w->y);
	}
	return PFPathClear(gstate);
}

Error PFStringShowKernJust(Gstate gstate, char *string, pt width)
{
	coor *w=&gstate->CharWidth;
	char *s=string;
	pt add=0,wdt; short len=0;
	short code, next;

	if (!string) return ERR_OK;

	/* determine length of string */
	while (*s) if (*s++==' ') len++;
	PFKernWidth(gstate,string,&wdt);
	if (len) add=(width-wdt)/len;

	code=PFChar2UniCode(gstate,string,&string);
	while (StringContinues(gstate,code))
	{
		PFShowChar(gstate, code);
		next=PFChar2UniCode(gstate,string,&string);
		KernAdjust(gstate, code, next);
		if (*string++==' ')
			PFMoveR(gstate, w->x+add, w->y);
		else
			PFMoveR(gstate, w->x, w->y);
		code=next;
	}
	return PFPathClear(gstate);
}

Error PFStringWidth(Gstate gstate, char *string, pt *xwidth)
{
	coor *w=&gstate->CharWidth;

	if (gstate->curr_info==0) return ERR_ITNF;
	*xwidth=0;

	if (!string) return ERR_OK;

	while (StringContinues(gstate,*string))
	{
		CharWidth(gstate, PFChar2UniCode(gstate,string,&string));
		*xwidth+=w->x;
	}
	return ERR_OK;
}

Error PFStringWidthKern(Gstate gstate, char *string, pt *xwidth)
{
	coor *w=&gstate->CharWidth;
	short code, next;

	if (gstate->curr_info==0) return ERR_ITNF;
	*xwidth=0;

	if (!string) return ERR_OK;

	code=PFChar2UniCode(gstate,string,&string);
	while (StringContinues(gstate,code))
	{
		CharWidth(gstate, code);
		next=PFChar2UniCode(gstate,string,&string);
		KernAdjust(gstate, code, next);
		*xwidth+=w->x;
		code=next;
	}
	return ERR_OK;
}
