/*
	PROforma, PROGS FOnt & Raster MAnager

	Hinting mechanism for PROforma fonts

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 22 February 1994

	VERSIONS :
	25.02.1994 first version completed
	19.03.1994 adjusted, because hinting can only work if scaled positive !
	12.10.1995 adjusted for better cache which also stores the hinting data
			   removed InitNoHintData
	18.10.1995 widthcoor now calculates width at fontsize
	21.10.1995 fixed bug in hintMatrix - started with currCTM io todefaultCTM
	28.10.1996 - added some output in case of CHARLOG
			   - stem aligning on grid inproved
			   - overshoot suppression was (very) bugged
			   - small problem with calculating stem snap values fixed
	30.10.1996 quick attempt to improve hinting by continuing to apply the
			   adjustment to all following coordinates when not in a stem
			   slight improvement, but not good enough
	31.10.1996 quick attempt to improve hinting by not just using a scale
			   factor with an adjustment, but also by scaling all points
			   differently in each zone, where zones are bounded by stem
			   boundaries -- fully working on 04.11.1996
	26.12.1996 hintMatrix now clears xd and yd in matrix
	06.04.1997 fixed a problem when calculating the hintmatrix when pixels
			   are not square
	18.06.1997 hinting now only applied when less than 50 pixels used for the
			   char (for 1000 units, in each direction separately)
*/

#include <array.h>
#include "gstate.h"
#include "system.h"
#include "cache.h"

#define any_abs(x) ((x)<0?-(x):(x))
#define Xdist(x)  ((x)*hd->matrix.xs)
#define Ydist(y) (-(y)*hd->matrix.ys)	/* y coor is reversed !! */
#define UNDEF 0xffff8000
#define HINT_LIMIT double2pt(50.0/1000.0)

#ifdef CHARLOG
#include <io.h>
extern Channel log;
#endif CHARLOG

static int Chint(const void *n1,const void *n2)
{
	return ((hint *)n1)->org-((hint *)n2)->org;
}

/* forward declarations */
void Compute_Zones(font_info *fi, hintdata *hd);
pt find_snap(short count, pt *snaps, pt width);
zone *find_zone(hintdata *hd, pt start, pt end);

void HintX(hintdata *hd, short x, short dx)
{
#ifdef CHARLOG
	char line[100];
#endif CHARLOG
	hint *h=&hd->Xhints[hd->XhintCount];
	pt width, hwidth, center, xpt;

	if (hd->useXhints && hd->XhintCount<MAXHINTS)
	{
		hd->XhintCount+=2;			/* there is one more hint */

		xpt=Xdist(x); width=Xdist(dx);
		if (width<0) { xpt+=width; width=-width; }
#ifdef CHARLOG
	STRFormatPut(line,"hintinfo x : start %f, width %f",pt2double(xpt),pt2double(width));
	IOPutString(log,line,NULL);
#endif CHARLOG
		hwidth=width/2;
		h[0].org=xpt	  -hd->BlueFuzz;
		h[1].org=xpt+width+hd->BlueFuzz;
		width=pt_round( find_snap(hd->XsnapCount,hd->XSnap,width) );

		if (width<pt_one) width=pt_one;
		if (hd->ForceBold && width<pt_one+pt_one) width=pt_one+pt_one;

		if ((width+pt_half)&pt_one)
			h[0].real=pt_center(xpt+hwidth)-(width/2);
		else
			h[0].real=pt_round (xpt+hwidth)-(width/2);
		h[1].real=h[0].real+width;
#ifdef CHARLOG
	STRFormatPut(line,"hintinfo x : new width %f, start %f, end %f",pt2double(width),pt2double(h[0].real),pt2double(h[1].real));
	IOPutString(log,line,NULL);
#endif CHARLOG
	}
	hd->calculated=FALSE;
}

void HintY(hintdata *hd, short y, short dy)
{
#ifdef CHARLOG
	char line[100];
#endif CHARLOG
	hint *h=&hd->Yhints[hd->YhintCount];
	pt width, hwidth, center, ypt;
	zone *z;

	if (hd->useYhints && hd->YhintCount<MAXHINTS)
	{
		hd->YhintCount+=2;			/* there is one more hint */

		ypt=Ydist(y); width=Ydist(dy);
		if (width<0) { ypt+=width; width=-width; }
#ifdef CHARLOG
	STRFormatPut(line,"hintinfo y : start %f, width %f",pt2double(ypt),pt2double(width));
	IOPutString(log,line,NULL);
#endif CHARLOG
		hwidth=width/2;
		h[0].org=ypt	  -hd->BlueFuzz;
		h[1].org=ypt+width+hd->BlueFuzz;

		z=find_zone(hd, ypt, ypt+width);

		width=pt_round( find_snap(hd->YsnapCount,hd->YSnap,width) );

		if (width<pt_one) width=pt_one;

		if (z)
		{
			/* adjust stem according to alignment zone !! */
			if (z->type==ZONE_TOP)
			{
				if ((z->end-h[0].org)>hd->BlueShift)
				{
					/* enforce overshoot */
					if ((h[0].real=pt_round(h[0].org))<pt_round(z->end)-pt_one)
						h[0].real-=pt_one;
				} else
					h[0].real=pt_round(z->end);
				h[1].real=h[0].real+width;
			} else {
				if ((h[1].org-z->start)>hd->BlueShift)
				{
					/* enforce overshoot */
					if ((h[1].real=pt_round(h[1].org))<pt_round(z->start)-pt_one)
						h[1].real+=pt_one;
				} else
					h[1].real=pt_round(z->start);
				h[0].real=h[1].real-width;
			}
		}
		else
		{
			if ((width+pt_half)&pt_one)
				h[0].real=pt_center(ypt+hwidth)-(width/2);
			else
				h[0].real= pt_round(ypt+hwidth)-(width/2);
			h[1].real=h[0].real+width;
		}
#ifdef CHARLOG
	STRFormatPut(line,"hintinfo y : new width %f, start %f, end %f",pt2double(width),pt2double(h[0].real),pt2double(h[1].real));
	IOPutString(log,line,NULL);
#endif CHARLOG
	}
	hd->calculated=FALSE;
}

void widthcoor(Gstate gstate, coor *c)
{
	register pt x=c->x, y=c->y;
	register CTM *m=(CTM *)gstate->currfont->info->FontMatrix;

	c->x = fixmul( x*m->xs - y*m->xr , gstate->fontsize);
	c->y = fixmul( x*m->yr - y*m->ys , gstate->fontsize);
}

void font2dev(hintdata *hd, coor *c)
{
#ifdef CHARLOG
	char line[100];
#endif CHARLOG
	pt x, y, t;
	int i; hint *h,*p;

	x = c->x*hd->matrix.xs - c->y*hd->matrix.xr;
	y = c->x*hd->matrix.yr - c->y*hd->matrix.ys;

	if (!hd->calculated)
	{
		ARRAYSort(hd->Xhints,hd->XhintCount,sizeof(hint),Chint);
		for (i=hd->XhintCount-1,h=hd->Xhints ; i>0 ; i--)
		{
			p=h++;
			/* the real coordinates should also be sorted, otherwise there */
			/* are two stems mapped on the same pixel --> handle that */
			if (h->real<p->real)
			{
				p->real=(p-1)->real;
				p->factor=0;
				h->real=(h+1)->real;
			}
			t=(h->org-p->org);
			h->factor=t ? fixdiv((h->real-p->real),t) : 0;
		}
#ifdef CHARLOG
for (i=hd->XhintCount,h=hd->Xhints ; i>0 ; i--,h++)
{
	STRFormatPutSafe(line,100,"hintlist x : old %f, new %f, factor %f",
		pt2double(h->org),pt2double(h->real),pt2double(h->factor));
	IOPutString(log,line,NULL);
}
#endif CHARLOG
		ARRAYSort(hd->Yhints,hd->YhintCount,sizeof(hint),Chint);
		for (i=hd->YhintCount-1,h=hd->Yhints ; i>0 ; i--)
		{
			p=h++;
			if (h->real<p->real)
			{
				p->real=(p-1)->real;
				p->factor=0;
				h->real=(h+1)->real;
			}
			t=(h->org-p->org);
			h->factor=t ? fixdiv((h->real-p->real),t) : 0;
		}
#ifdef CHARLOG
for (i=hd->YhintCount,h=hd->Yhints ; i>0 ; i--,h++)
{
	STRFormatPutSafe(line,100,"hintlist y : old %f, new %f, factor %f",
		pt2double(h->org),pt2double(h->real),pt2double(h->factor));
	IOPutString(log,line,NULL);
}
#endif CHARLOG
		hd->calculated=TRUE;
	}

	if (hd->XhintCount)
	{
		h=hd->Xhints;
		if (x<h->org)
			x+=h->real-h->org;
		else
		{
			for (i=hd->XhintCount-1 ; i ; i--)
			{
				h++;
				if (x<h->org)
				{
					p=h-1;
					x=p->real+fixmul(x-p->org,h->factor);
					goto Xdone;
				}
			}
			x+=h->real-h->org;
		}
	}
Xdone:
	if (hd->YhintCount)
	{
		h=hd->Yhints;
		if (y<h->org)
			y+=h->real-h->org;
		else
		{
			for (i=hd->YhintCount-1 ; i ; i--)
			{
				h++;
				if (y<h->org)
				{
					p=h-1;
					y=p->real+fixmul(y-p->org,h->factor);
					goto Ydone;
				}
			}
			y+=h->real-h->org;
		}
	}
Ydone:
	c->x=x;
	c->y=y;
}

Error hintMatrix(Gstate gstate, CTM *fm, CTM *matrix)
{
	CTM *cm=&gstate->todefaultCTM;
	pt scale;

	/* calculate matrix which identifies the required CachedFont */
	/* by applying FontMatrix on current CTM */
	/* & make sure this scales to fontsize and converts to device coor */

	scale=fixmul(gstate->todev.x,gstate->fontsize);
	matrix->xs=fixmul( fixmul(cm->xs, fm->xs)+fixmul(cm->yr,-fm->xr) ,scale);
	matrix->xr=fixmul( fixmul(cm->xr, fm->xs)+fixmul(cm->ys,-fm->xr) ,scale);
	scale=fixmul(gstate->todev.y,gstate->fontsize);
	matrix->ys=fixmul( fixmul(cm->ys, fm->ys)+fixmul(cm->xr,-fm->yr) ,scale);
	matrix->yr=fixmul( fixmul(cm->xs,-fm->yr)+fixmul(cm->yr, fm->ys) ,scale);
	matrix->xd=matrix->yd=0;

	return ERR_OK;
}

/* Initialise hint data for character drawing (only used by PROforma) */
void InitFontHints(hintdata *hd, font_info *fi)
{
	short i;

	/* start interpretting the actual hints !!! */

	hd->useXhints=hd->useYhints=FALSE;
	if (hd->matrix.xr==0 && hd->matrix.xs>0 && hd->matrix.xs<HINT_LIMIT)
		hd->useXhints=TRUE;
	if (hd->matrix.yr==0 && hd->matrix.ys>0 && hd->matrix.ys<HINT_LIMIT)
		hd->useYhints=TRUE;

	if (hd->useXhints)
	{
		/* calculate snapping values */
		hd->XsnapCount=0;
		if (fi->StdVW!=UNDEF)
			hd->XSnap[hd->XsnapCount++]=any_abs(Xdist(fi->StdVW));
		for (i=0;i<12 && fi->StemSnapV[i]!=UNDEF;i++)
			hd->XSnap[hd->XsnapCount++]=any_abs(Xdist(fi->StemSnapV[i]));
	}
	if (hd->useYhints)
	{
		/* calculate snapping values */
		hd->YsnapCount=0;
		if (fi->StdHW!=UNDEF)
			hd->YSnap[hd->YsnapCount++]=any_abs(Ydist(fi->StdHW));
		for (i=0;i<12 && fi->StemSnapH[i]!=UNDEF;i++)
			hd->YSnap[hd->YsnapCount++]=any_abs(Ydist(fi->StemSnapH[i]));

		/* calculate BlueValues */

		/* calculate BlueShift, should be at least half a pixel when used (section 5.7) */
		hd->BlueShift=any_abs(Ydist(fi->BlueShift));
		if (hd->BlueShift<pt_half) hd->BlueShift=pt_half;
		if (Ydist(1)<fi->BlueScale) hd->BlueShift=ptmax/2;

		/* fill in BlueFuzz and ForceBold */
		hd->BlueFuzz=any_abs(Ydist(fi->BlueFuzz));
		hd->ForceBold=FALSE;

		Compute_Zones(fi,hd);
	}
}

#define setblues if (start<end) blues->start=start-BlueFuzz, blues->end=end+BlueFuzz; \
						   else blues->start=end-BlueFuzz, blues->end=start+BlueFuzz

void Compute_Zones(font_info *fi, hintdata *hd)
{
	short i;
	zone *blues=hd->BlueValues;
	pt BlueFuzz=fi->BlueFuzz;
	pt start, end;

	hd->BlueCount=0;
	if (fi->BlueValues[0]!=UNDEF && fi->BlueValues[1]!=UNDEF)
	{
		blues->type =ZONE_BOTTOM;
		start=Ydist(fi->BlueValues[0]);
		end  =Ydist(fi->BlueValues[1]);
		setblues;

		/* check if Family values should supersede */
		if (fi->FamilyBlues[0]!=UNDEF && fi->FamilyBlues[1]!=UNDEF)
		{
			start=Ydist(fi->FamilyBlues[0]);
			end  =Ydist(fi->FamilyBlues[1]);
			if ((start<end && any_abs(blues->end-end)<pt_one) || any_abs(blues->end-start)<pt_one)
			{ setblues; }
		}
		blues++;
		hd->BlueCount++;
	}
	for (i=2;i<10 && fi->BlueValues[i]!=UNDEF && fi->BlueValues[i+1]!=UNDEF; i+=2)
	{
		blues->type =ZONE_TOP;
		start=Ydist(fi->BlueValues[i]);
		end  =Ydist(fi->BlueValues[i+1]);
		setblues;

		/* check if Family values should supersede */
		if (fi->FamilyBlues[i]!=UNDEF && fi->FamilyBlues[i+1]!=UNDEF)
		{
			start=Ydist(fi->FamilyBlues[i]);
			end  =Ydist(fi->FamilyBlues[i+1]);
			if ((start<end && any_abs(blues->end-end)<pt_one) || any_abs(blues->end-start)<pt_one)
			{ setblues; }
		}

		blues++;
		hd->BlueCount++;
	}
	for (i=0;i<10 && fi->OtherBlues[i]!=UNDEF && fi->OtherBlues[i+1]!=UNDEF; i+=2)
	{
		blues->type =ZONE_BOTTOM;
		start=Ydist(fi->OtherBlues[i]);
		end  =Ydist(fi->OtherBlues[i+1]);
		setblues;

		/* check if Family values should supersede */
		if (fi->FamilyOtherBlues[i]!=UNDEF && fi->FamilyOtherBlues[i+1]!=UNDEF)
		{
			start=Ydist(fi->FamilyOtherBlues[i]);
			end  =Ydist(fi->FamilyOtherBlues[i+1]);
			if ((start<end && any_abs(blues->end-end)<pt_one) || any_abs(blues->end-start)<pt_one)
			{ setblues; }
		}

		blues++;
		hd->BlueCount++;
	}
}

pt find_snap(short count, pt *snaps, pt width)
{
	short i; pt snap=width,delta=pt_half;

	for (i=count;i;i--,snaps++)
	{
		if (any_abs(width-*snaps)<delta)
			delta=any_abs(width-*snaps), snap=*snaps;
	}
	return snap;
}

zone *find_zone(hintdata *hd, pt start, pt end)
{
	short i; zone *z=hd->BlueValues;

	for (i=hd->BlueCount;i;i--,z++)
	{
		if (z->type==ZONE_TOP)
		{
			if (start>=z->start && start<=z->end)
				return z;
		}
		else /* ZONE_BOTTOM */
		{
			if (end>=z->start && end<=z->end)
				return z;
		}
	}
	return NULL;
}
