/*
	PROforma, PROgs FOnt & Raster MAnager

	example picture driver

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	on 21 December 1995
	can freely by used as a template for PROforma picture drivers.
*/

#include <io.h>
#include "PFmodule.h"

typedef struct {
	char identifier[24];	/* "example picture format" */
	short xsiz, ysiz;		/* size in pixels */
	short xratio, yratio;	/* pixel aspect ratio (0 if not known) */
	unsigned char data[2];	/* start of picture data */
} Picture;

Error PFPictureElement(Gstate gstate, pt xsiz, pt ysiz, pt xorg, pt yorg);

static Error Handle(int command,...)
{
	return ERR_NIMP;
}

static Error Recognize(Gstate gstate, char *base)
{
	FileInfo *fi=(FileInfo *)base;
	Picture *pict=(Picture *)(base+sizeof(FileInfo));

	if (fi->type==FILETYPE_NORMAL &&
		STRSameCD(pict->identifier,"example picture format") &&
		pict->xsiz>0 && pict->ysiz>0)
		return ERR_OK;
	else
		return ERR_INAM;
}

static Error AspectRatio(Gstate gstate, char *base, int *xratio, int *yratio)
{
	Picture *pict=(Picture *)(base+sizeof(FileInfo));
	if (pict->xratio && pict->yratio)
	{
		*xratio=pict->xratio*pict->xsiz;
		*yratio=pict->yratio*pict->ysiz;
		return ERR_OK;
	}
	else
	{
		*xratio=4; *yratio=3;
		return ERR_ITNF;
	}
}

static Error ColourInfo(Gstate gstate, char *base, int *count, int *space)
{
	*count=256;
	*space=PF_COLOURSPACE_RGB;
	return ERR_OK;
}

static Error ColourGet(Gstate gstate, char *base, int which, void *colour)
{
	/* the colours are 8 bit : iirrggbb (i for intensity) */
	ColourRGB *rgb=(ColourRGB *)colour;
	int intensity=((which>>6)&3)+1;

	rgb->red  =intensity * ((which>>4)&3) * (pt_hundred/12);
	rgb->green=intensity * ((which>>2)&3) * (pt_hundred/12);
	rgb->blue =intensity * ((which	 )&3) * (pt_hundred/12);
	return ERR_OK;
}

static Error Display(Gstate gstate, char *base, pt xsiz, pt ysiz,
							Error (*ColourSelect)(Gstate, int))
{
	Picture *pict=(Picture *)(base+sizeof(FileInfo));
	int xpix,ypix=pict->ysiz;
	int colour, backcolour=0;
	short bit=0x80;
	int length, lastcolour;
	unsigned char *linestart=pict->data;
	unsigned char *linepos;
	int posx;
	pt posy=0;

	/* the background colour is random (well 0) - actually choosing */
	/* the background colour would require running through the entire */
	/* picture once more -> expensive */

	/* first check whether some part of the picture is visible ! */
	ColourSelect(gstate,backcolour);
	if (PFPictureElement(gstate,xsiz,ysiz,0,posy)) return ERR_OK;

	xsiz/=pict->xsiz;
	ysiz/=pict->ysiz;

	while(ypix)
	{
		posx=0;
		linepos=linestart;

		/* see if this line is clipped - and set background colour */
		ColourSelect(gstate,backcolour);
		if (!PFPictureElement(gstate,xsiz*pict->xsiz,ysiz,0,posy))
		{
			length=1;
			/* get colour */
			lastcolour=*linepos;

			for(xpix=pict->xsiz-1; xpix; xpix--)
			{
				/* get colour */
				colour=*linepos++;

				if (colour!=lastcolour)
				{
					if (lastcolour!=backcolour)
					{
						ColourSelect(gstate,lastcolour);
						PFPictureElement(gstate,xsiz*length,ysiz,xsiz*posx,posy);
					}
					/* skip pixels on page */
					posx+=length;
					lastcolour=colour;
					length=1;
				} else
					length++;
			}
			/* there may be a sequence left at the end of the line */
			if (colour!=backcolour)
			{
				ColourSelect(gstate,colour);
				PFPictureElement(gstate,xsiz*length,ysiz,xsiz*posx,posy);
			}
		}
		posy+=ysiz;
		ypix--;
		linestart+=pict->xsiz;
	}
	return ERR_OK;
}

PICTdriver init = {
	NULL, PF_PICTUREDRIVER,
	"example 256 colour picture",

	/* can we recognize the picture */
	Recognize,
	AspectRatio,

	/* display the picture */
	Display,

	/* info on colours used in the picture */
	ColourInfo,
	ColourGet,

	Handle
};
